// SPDX-License-Identifier: GPL-2.0+
/*
 * linux/fs/erofs/unzip.c
 *
 * Copyright (c) 2018 HUAWEI, Inc.
 *             http://www.huawei.com/
 * Created by Gao Xiang <gaoxiang25@huawei.com>
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file COPYING in the main directory of the Linux
 * distribution for more details.
 */
#include "unzip.h"
#include <linux/slab.h>
#include <linux/blkdev.h>
#include <linux/prefetch.h>

static struct workqueue_struct *z_erofs_workqueue __read_mostly;
static struct kmem_cache *z_erofs_pack_cachep __read_mostly;

void z_erofs_exit_zip_subsystem(void)
{
	BUG_ON(z_erofs_workqueue == NULL);
	BUG_ON(z_erofs_pack_cachep == NULL);

	destroy_workqueue(z_erofs_workqueue);
	kmem_cache_destroy(z_erofs_pack_cachep);
}

static inline int init_unzip_workqueue(void)
{
	const unsigned onlinecpus = num_online_cpus();

	/*
	 * we don't need too many threads, limiting threads
	 * could improve scheduling performance.
	 */
	z_erofs_workqueue = alloc_workqueue("erofs_unzipd",
		WQ_UNBOUND | WQ_CPU_INTENSIVE | WQ_HIGHPRI |
		WQ_NON_REENTRANT, onlinecpus + onlinecpus / 4);

	return z_erofs_workqueue != NULL ? 0 : -ENOMEM;
}

int z_erofs_init_zip_subsystem(void)
{
	z_erofs_pack_cachep =
		kmem_cache_create("erofs_compressed_pack",
		Z_EROFS_PACK_SIZE, 0,
		SLAB_RECLAIM_ACCOUNT, NULL);

	if (z_erofs_pack_cachep != NULL) {
		if (!init_unzip_workqueue())
			return 0;

		kmem_cache_destroy(z_erofs_pack_cachep);
	}
	return -ENOMEM;
}

static inline void put_vle_zipped_pack(struct z_erofs_vle_zipped_pack *z,
                                       bool __maybe_unused allow_free)
{
	if (erofs_put_page_bundle(&z->bundle))
		return;

	DBG_BUGON(mutex_is_locked(&z->lock));
	DBG_BUGON(!allow_free);
	kmem_cache_free(z_erofs_pack_cachep, z);
}

int erofs_try_to_free_vle_zipped_page(struct page *page)
{
	struct erofs_page_bundle *b;
	struct z_erofs_vle_zipped_pack *zip;
	unsigned i;
	bool will_free;

	erofs_dbg_might_sleep();
	b = erofs_lock_page_private(page);

	DBG_BUGON(!has_page_bundle(page));
	zip = container_of(b, struct z_erofs_vle_zipped_pack, bundle);

	/* I prefer not to sleep in the reclaim path, try_lock instead */
	if (!mutex_trylock(&zip->lock)) {
busy_unlock_page_private:
		erofs_unlock_page_private(page);
		return 0;
	}

	/* freeze the whole page bundle */
	spin_lock(&b->lockref.lock);

	/* the page bundle still has active users */
	if (b->lockref.count > 1) {
busy_unlock_bundle:
		spin_unlock(&b->lockref.lock);
		mutex_unlock(&zip->lock);
		goto busy_unlock_page_private;
	}

	/* try to release the head zipped page */
	if (page == b->pages[0]) {
		/* the rest zpages should be released */
		for(i = 1; i < EROFS_PAGE_BUNDLE_MAX_PAGES; ++i)
			if (b->pages[i] != NULL)
				goto busy_unlock_bundle;
		b->pages[0] = NULL;
		will_free = true;
		goto reclaim;
	}

	for(i = 1; i < EROFS_PAGE_BUNDLE_MAX_PAGES; ++i) {
		if (b->pages[i] == page) {
			b->pages[i] = NULL;
			will_free = false;
			goto reclaim;
		}
	}

	BUG();
reclaim:
	ClearPagePrivate(page);
	erofs_set_page_private(page, NULL);
	spin_unlock(&b->lockref.lock);
	mutex_unlock(&zip->lock);
	erofs_unlock_page_private(page);

	if (will_free)
		put_vle_zipped_pack(zip, true);
	put_page(page);
	return 1;
}

/* zip should be locked by callers */
static void z_erofs_vle_unzip(struct z_erofs_vle_zipped_pack *const zip)
{
	struct erofs_page_bundle *const b = &zip->bundle;
	struct z_erofs_pack_info pack;
	struct inode *inode;
	struct page *page;
#if EROFS_PAGE_BUNDLE_MAX_PAGES > 1
	unsigned clusterpages, i;
#else
	const unsigned clusterpages = 1;
#endif
	void *in;

	/*
	 * end_io queue work start
	 * end_io work queue end (queued_pages == 0)
	 * z_erofs_vle_do_read_page, queue work again
	 */
	if (unlikely(!READ_ONCE(zip->queued_pages)))
		goto out_unlock;

	page = zip->pages[0];
	DBG_BUGON(page == NULL);
	inode = page->mapping->host;

#if EROFS_PAGE_BUNDLE_MAX_PAGES > 1
	clusterpages = erofs_clusterpages(EROFS_I_SB(inode));

	for(i = 0; i < clusterpages; ++i) {
		DBG_BUGON(b->pages[i] == NULL);
		DBG_BUGON(!PageUptodate(b->pages[i]));
	}
#else
	DBG_BUGON(b->pages[0] == NULL);
	DBG_BUGON(!PageUptodate(b->pages[0]));
#endif

	debugln("%s, zip=%p la = %llu, llen = %u", __func__, zip, zip->la, zip->llen);

	pack.pages = zip->pages;
	pack.nr_pages = zip->nr_pages;
	pack.queued_pages = zip->queued_pages;

	if (!(zip->flags & Z_EROFS_PACK_ZIPPED))
		z_erofs_plain_copy(&pack, b->pages, clusterpages, zip->la);
	else {
#if EROFS_PAGE_BUNDLE_MAX_PAGES > 1
		in = clusterpages == 1 ? kmap(b->pages[0]):
			vmap(b->pages, clusterpages, VM_MAP, PAGE_KERNEL);
#else
		in = kmap(b->pages[0]);
#endif

		z_erofs_unzip_generic(&pack, in, clusterpages * PAGE_SIZE,
		                      zip->la, zip->llen);
#if EROFS_PAGE_BUNDLE_MAX_PAGES > 1
		if (clusterpages == 1)
			kunmap(b->pages[0]);
		else
			vunmap(in);
#else
		kunmap(b->pages[0]);
#endif
	}

	/* check decompressor has filled all queued pages */
	DBG_BUGON(pack.queued_pages);
	zip->queued_pages = 0;
	zip->nr_pages = 0;		/* FIXME later */
out_unlock:
	mutex_unlock(&zip->lock);
	put_vle_zipped_pack(zip, false);

}

static void z_erofs_vle_decompress_wq(struct work_struct *work)
{
	struct z_erofs_vle_zipped_pack *const zip =
		container_of(work, struct z_erofs_vle_zipped_pack, work);

	if (!READ_ONCE(zip->queued_pages)) {
		put_vle_zipped_pack(zip, false);
		return;
	}
	mutex_lock(&zip->lock);
	z_erofs_vle_unzip(zip);
}

static void __vle_zipped_bundle_alloc(struct page *page, unsigned nr)
{
	struct erofs_page_bundle *b;
	struct z_erofs_vle_zipped_pack *zip =
		kmem_cache_zalloc(z_erofs_pack_cachep, GFP_ATOMIC);

	/* here we grab an extra page reference for page private */
	get_page(page);

	/* if we cannot allocate memory in atomic, try sleeping way instead */
	if (unlikely(zip == NULL)) {
		erofs_unlock_page_private(page);

		erofs_dbg_might_sleep();
		zip = kmem_cache_zalloc(z_erofs_pack_cachep,
		                        GFP_KERNEL | __GFP_NOFAIL);

		b = erofs_lock_page_private(page);
		if (test_set_page_bundle(page)) {
			DBG_BUGON(b == NULL);
			DBG_BUGON(b->pages[nr] != page);

			lockref_get(&b->lockref);
			kmem_cache_free(z_erofs_pack_cachep, zip);
			put_page(page);
			return;
		}

		DBG_BUGON(b != NULL);
	} else if (test_set_page_bundle(page))
		BUG();

	mutex_init(&zip->lock);
	INIT_WORK(&zip->work, z_erofs_vle_decompress_wq);

	b = &zip->bundle;
	/* initialize global page bundle */
	b->pages[nr] = page;
	b->lockref.count = 2;
	spin_lock_init(&b->lockref.lock);
	erofs_set_page_private(page, b);
}

static inline struct page *grab_vle_zipped_page(struct super_block *sb,
                                                pgoff_t index,
                                                struct erofs_page_bundle **b,
                                                bool *created,
                                                struct list_head *page_pool)
{
	struct page *page;

	page = erofs_grab_bundle_page(sb, index, created, page_pool);
	if (!IS_ERR(page)) {
		/* we only get a new page bundle from the head page */
		*b = erofs_get_page_bundle(page, 0, __vle_zipped_bundle_alloc);
	}
	return page;
}

/* TODO! FIXME!!! this function is still broken :( */
static int z_erofs_add_tailpage(struct z_erofs_zipped_pagevec *z_pvec,
                                struct super_block *sb,
                                pgoff_t hi, pgoff_t ti,
                                struct erofs_page_bundle *b,
                                struct list_head *page_pool)
{
	return -ENOTSUPP;
}

struct z_erofs_zipped_pack_collector {
	struct list_head list;
	bool sync;
};

static inline void vle_zipped_iter_dispatch(struct z_erofs_vle_zipped_iter *z,
	struct z_erofs_zipped_pack_collector *c)
{
	struct z_erofs_vle_zipped_pack *const zip = z->zip;
	struct list_head *const e = z_erofs_vle_zipped_list_entry(zip);

	/* decompressed pages is already ok? */
	if (!z->already) {
		if (c->sync) {
			if (!z_erofs_vle_zipped_protect_list_entry(zip))
				return;
			list_add_tail(e, &c->list);
		}
	} else {
		if (!z_erofs_vle_zipped_protect_list_entry(zip))
			return;
		list_add(e, &c->list);
	}
	lockref_get(&zip->bundle.lockref);
}

static inline void vle_zipped_iter_end(struct z_erofs_vle_zipped_iter *z)
{
	z_erofs_de_pagevec_end(&z->d_pvec, false);
	mutex_unlock(&z->zip->lock);

	put_vle_zipped_pack(z->zip, false);
}

static inline void vle_zipped_collected_enqueue_all(struct list_head *list)
{
	struct list_head *e, *tmp;

	list_for_each_safe(e, tmp, list) {
		struct work_struct *work = container_of(e,
			struct work_struct, entry);
		struct z_erofs_vle_zipped_pack *zip;

		list_del(e);
		INIT_LIST_HEAD(e);

		zip = container_of(work, struct z_erofs_vle_zipped_pack, work);
		z_erofs_vle_zipped_unprotect_list_entry(zip);

		/* there is no need to lock strictly */
		if (unlikely(!READ_ONCE(zip->queued_pages))) {
			put_vle_zipped_pack(zip, false);
			continue;
		}
		debugln("%s, queue work %p", __func__, &zip->work);
		queue_work(z_erofs_workqueue, work);
	}
}

static inline void vle_zipped_collected_unzip_all(struct super_block *sb,
	struct list_head *list)
{
	struct work_struct *work;
	struct z_erofs_vle_zipped_pack *zip;
	struct erofs_page_bundle *b;
	struct page *victim;
#if EROFS_PAGE_BUNDLE_MAX_PAGES > 1
	unsigned clusterpages = erofs_clusterpages(EROFS_SB(sb));
	unsigned j;
#endif
	struct list_head *e, *tmp;
	unsigned round = 0;

repeat:
	/* why isn't blk_flush_plug_list() exported? :-( */
	if (round == 1 && blk_needs_flush_plug(current))
		io_schedule();

	/* wait on a single page at each end of a round */
	victim = NULL;

	list_for_each_safe(e, tmp, list) {
		work = container_of(e, struct work_struct, entry);
		zip = container_of(work, struct z_erofs_vle_zipped_pack, work);
		b = &zip->bundle;

#if EROFS_PAGE_BUNDLE_MAX_PAGES > 1
		for (j = 0; j < clusterpages; ++j) {
			if (!PageLocked(b->pages[j]))
				continue;
			if (round >= 4)
				if (victim == NULL || !PageLocked(victim))
					victim = b->pages[j];
			break;
		}
		if (j < clusterpages) {
#else
		if (PageLocked(b->pages[0])) {
			if (victim == NULL || !PageLocked(victim))
				victim = b->pages[0];
#endif
			continue;
		}

#if EROFS_PAGE_BUNDLE_MAX_PAGES > 1
		for (j = 0; j < clusterpages; ++j)
			BUG_ON(!PageUptodate(b->pages[j]));
#else
		BUG_ON(!PageUptodate(b->pages[0]));
#endif

		if (round >= 6)
			mutex_lock(&zip->lock);
		else if (!mutex_trylock(&zip->lock))
			continue;

		list_del(e);
		INIT_LIST_HEAD(e);
		z_erofs_vle_zipped_unprotect_list_entry(zip);
		z_erofs_vle_unzip(zip);
	}

	if (!list_empty(list)) {
		if (victim != NULL)
			wait_on_page_locked(victim);

		++round;
		goto repeat;
	}
}

static int z_erofs_vle_do_read_page(
	struct page *page,
	struct z_erofs_zipped_pagevec *z_pvec,
	struct z_erofs_vle_zipped_iter *z,
	struct erofs_map_blocks_iter *m,
	struct list_head *page_pool,
	struct z_erofs_zipped_pack_collector *collector)
{
	struct inode *const inode = page->mapping->host;
	struct super_block *const sb = inode->i_sb;
	struct erofs_sb_info *const sbi = EROFS_SB(sb);
	const loff_t offset = page_offset(page);
	struct z_erofs_vle_zipped_pack *zip = z->zip;
	unsigned cur, end, spiltted;
	int err;
	bool creat;
	struct page *zpage;
	struct erofs_page_bundle *b;
	unsigned clusterpages;
	pgoff_t hi, ti;

	/* register locked file pages as online pages in pack */
	z_erofs_onlinepage_init(page);

	spiltted = 0;
	end = PAGE_SIZE;
repeat:
	cur = end - 1;

	/* lucky, within the range of the current map_blocks */
	if (offset + cur >= m->map.m_la &&
            offset + cur < m->map.m_la + m->map.m_llen)
		goto hitted;

	/* go ahead the next map_blocks */
	debugln("%s: [out-of-range] pos %llu", __func__, offset + cur);

	if (zip != NULL) {
		vle_zipped_iter_dispatch(z, collector);
		vle_zipped_iter_end(z);
	}

	m->map.m_la = offset + cur;
	m->map.m_llen = 0;
	err = erofs_map_blocks_iter(inode, &m->map, &m->mpage, 0);
	if (unlikely(err))
		goto err_out;

	/* deal with hole (FIXME! broken now) */
	if (unlikely(!(m->map.m_flags & EROFS_MAP_MAPPED))) {
		zip = NULL;
		goto hitted;
	}

	DBG_BUGON(m->map.m_plen != 1 << sbi->clusterbits);
	BUG_ON(m->map.m_pa % EROFS_BLKSIZ);

	/* grab the zipped head page and bundle */
	hi = m->map.m_pa / PAGE_SIZE;
	zpage = grab_vle_zipped_page(sb, hi, &b, &creat, page_pool);

	zip = container_of(b, struct z_erofs_vle_zipped_pack, bundle);
	if (IS_ERR(zpage))
		goto err_out;

	debugln("%s, (head zipped page %p, index=%lu) page %p "
		"created=%d", __func__, zpage, hi, page, creat);

	clusterpages = erofs_clusterpages(sbi);

	/* already = true iff no zpage adds to zipped_pagevec */
	z->already = true;

	/* as others above, add tail zpages in the reserve order */
	ti = DIV_ROUND_UP(m->map.m_pa + m->map.m_plen, PAGE_SIZE);
	while(ti > hi + 1) {
		err = z_erofs_add_tailpage(z_pvec, sb, hi, --ti, b, page_pool);
		z->already &= !err;
	}

	if (!creat) {
		/* why do this? -- see comment in "do_read_cache_page" */
		wait_on_page_locked(zpage);

		if (PageUptodate(zpage))
			goto has_data;

		lock_page(zpage);
		if (PageUptodate(zpage)) {
			unlock_page(zpage);
			goto has_data;
		}
	}

	z_erofs_zipped_pagevec_push(z_pvec, zpage);
	z->already = false;

has_data:
	mutex_lock(&zip->lock);

	z->zip = zip;

	if (!(zip->flags & Z_EROFS_PACK_INITIALIZED)) {
		zip->la = m->map.m_la;
		if (m->map.m_flags & EROFS_MAP_ZIPPED)
			zip->flags |= Z_EROFS_PACK_ZIPPED;
		zip->flags |= Z_EROFS_PACK_INITIALIZED;
	} else {
		BUG_ON(zip->la != m->map.m_la);
		BUG_ON(!(zip->flags & Z_EROFS_PACK_ZIPPED) !=
			!(m->map.m_flags & EROFS_MAP_ZIPPED));
	}

	/* physical address should be equal */
	DBG_BUGON(m->map.m_pa != page_offset(b->pages[0]));

	/* update logical extent length */
	if (m->map.m_llen > zip->llen)
		zip->llen = m->map.m_llen;

	put_page(zpage);
	z_erofs_de_pagevec_init(&z->d_pvec, zip->pages, zip->queued_pages);

hitted:
	cur = end - min_t(unsigned, offset + end - m->map.m_la, end);
	if (unlikely(!(m->map.m_flags & EROFS_MAP_MAPPED))) {
		zero_user_segment(page, cur, end);
		goto next_part;
	}

	++spiltted;
	z_erofs_de_pagevec_enqueue(&z->d_pvec, page);

	/* also update nr_pages and increase queued_pages */
	zip->nr_pages = max_t(pgoff_t, zip->nr_pages,
	                      page->index - m->map.m_la / PAGE_SIZE + 1);
	++zip->queued_pages;

next_part:
	/* used for verification */
	m->map.m_llen = offset + cur - m->map.m_la;

	if ((end = cur) > 0)
		goto repeat;

	debugln("%s, finish page: %p spiltted: %u map->m_llen %llu",
		__func__, page, spiltted, m->map.m_llen);

	/* the online file page could be unlocked after this line */
	z_erofs_onlinepage_setup(page, spiltted);
	return 0;

err_out:
	/* TODO! the missing error handing cases */
	return err;
}


/*
 * Variable-sized Logical Extent (Fixed Physical Cluster) Compression Mode
 * ---
 * VLE compression mode attempts to compress a number of logical data into
 * a physical cluster with a fixed size.
 * VLE compression mode uses "struct erofs_decompressed_index_vle".
 */
static erofs_off_t vle_get_logical_extent_head(
	struct inode *inode,
	struct page **page_iter,
	void **kaddr_iter,
	unsigned lcn,	/* logical cluster number */
	erofs_blk_t *pcn,
	unsigned *flags)
{
	/* for extent meta */
	struct page *page = *page_iter;
	erofs_blk_t blkaddr = vle_extent_blkaddr(inode, lcn);
	struct erofs_decompressed_index_vle *di;
	unsigned long long ofs;
	unsigned clustersize = 1 << EROFS_SB(inode->i_sb)->clusterbits;

	if (page->index != blkaddr) {
		kunmap_atomic(*kaddr_iter);
		unlock_page(page);
		put_page(page);

		*page_iter = page = erofs_get_meta_page(inode->i_sb,
			blkaddr, false);
		*kaddr_iter = kmap_atomic(page);
	}

	di = *kaddr_iter + vle_extent_blkoff(inode, lcn);
	switch(vle_cluster_type(di)) {
	case EROFS_VLE_CLUSTER_TYPE_NONHEAD:
		BUG_ON(!di->di_u.delta[0]);
		BUG_ON(lcn < di->di_u.delta[0]);

		ofs = vle_get_logical_extent_head(inode,
			page_iter, kaddr_iter,
			lcn - di->di_u.delta[0], pcn, flags);
		break;
	case EROFS_VLE_CLUSTER_TYPE_PLAIN:
		*flags ^= EROFS_MAP_ZIPPED;
	case EROFS_VLE_CLUSTER_TYPE_HEAD:
		ofs = lcn * clustersize +
			(le16_to_cpu(di->di_clusterofs) & (clustersize - 1));
		*pcn = le32_to_cpu(di->di_u.blkaddr);
		break;
	default:
		BUG_ON(1);
	}
	return ofs;
}

int erofs_map_blocks_iter(struct inode *inode,
	struct erofs_map_blocks *map,
	struct page **mpage_ret, int flags)
{
	/* logicial extent (start, end) offset */
	unsigned long long ofs, end;
	struct erofs_decompressed_index_vle *di;
	erofs_blk_t e_blkaddr, pcn;
	unsigned lcn, logical_cluster_ofs;
	struct page *mpage = *mpage_ret;
	void *kaddr;
	bool initial;
	unsigned clustersize = 1 << EROFS_SB(inode->i_sb)->clusterbits;

	/* if both m_(l,p)len are 0, regularize l_lblk, l_lofs, etc... */
	initial = !map->m_llen;

	if (unlikely(map->m_la >= inode->i_size)) {
		BUG_ON(!initial);
		map->m_la = inode->i_size - 1;
	}

	debugln("%s, m_la %llu m_llen %llu --- start", __func__,
		map->m_la, map->m_llen);

	ofs = map->m_la + map->m_llen;

	lcn = ofs / clustersize;
	e_blkaddr = vle_extent_blkaddr(inode, lcn);

	if (mpage == NULL || mpage->index != e_blkaddr) {
		if (mpage != NULL)
			put_page(mpage);

		mpage = erofs_get_meta_page(inode->i_sb, e_blkaddr, false);
		*mpage_ret = mpage;
	} else {
		lock_page(mpage);
		DBG_BUGON(!PageUptodate(mpage));
	}

	kaddr = kmap_atomic(mpage);
	di = kaddr + vle_extent_blkoff(inode, lcn);

	debugln("%s, lcn %u e_blkaddr %u e_blkoff %u", __func__, lcn,
		e_blkaddr, vle_extent_blkoff(inode, lcn));

	logical_cluster_ofs = vle_compressed_index_clusterofs(clustersize, di);
	if (!initial) {
		/* m_(l,p)blk, m_(l,p)ofs has been already initialized */
		map->m_llen += logical_cluster_ofs;
		goto out;
	}

	/* by default, compressed */
	map->m_flags |= EROFS_MAP_ZIPPED;

	end = (u64)(lcn + 1) * clustersize;

	switch(vle_cluster_type(di)) {
	case EROFS_VLE_CLUSTER_TYPE_PLAIN:
		if (ofs % clustersize >= logical_cluster_ofs)
			map->m_flags ^= EROFS_MAP_ZIPPED;
	case EROFS_VLE_CLUSTER_TYPE_HEAD:
		if (ofs % clustersize == logical_cluster_ofs) {
			pcn = le32_to_cpu(di->di_u.blkaddr);
			goto unneed;
		}

		if (ofs % clustersize > logical_cluster_ofs) {
			ofs = lcn * clustersize | logical_cluster_ofs;
			pcn = le32_to_cpu(di->di_u.blkaddr);
			break;
		}

		BUG_ON(!lcn);	/* logical cluster number >= 1 */
		end = (lcn-- * clustersize) | logical_cluster_ofs;
	case EROFS_VLE_CLUSTER_TYPE_NONHEAD:
		/* get the correspoinding first chunk */
		ofs = vle_get_logical_extent_head(inode, mpage_ret,
			&kaddr, lcn, &pcn, &map->m_flags);
		mpage = *mpage_ret;
		break;
	default:
		errln("%s, invalid cluster type %u on m_la %llu of nid %llu",
			__func__, vle_cluster_type(di), ofs,
			EROFS_V(inode)->nid);
		BUG();
		pcn = ~0;
	}

	map->m_la = ofs;
unneed:
	map->m_llen = end - ofs;
	map->m_plen = clustersize;
	map->m_pa = blknr_to_addr(pcn);
	map->m_flags |= EROFS_MAP_MAPPED;
	debugln("%s, m_la %llu m_pa %llu m_llen %llu m_plen %llu m_flags %u",
		__func__, map->m_la, map->m_pa,
		map->m_llen, map->m_plen, map->m_flags);
out:
	kunmap_atomic(kaddr);
	unlock_page(mpage);
	return 0;
}

#if (LINUX_VERSION_CODE < KERNEL_VERSION(4, 3, 0))
static inline void zipped_async_read_endio(struct bio *bio, int err)
#else
static inline void zipped_async_read_endio(struct bio *bio)
#endif
{
#if EROFS_PAGE_BUNDLE_MAX_PAGES > 1
	struct super_block *sb = bio->bi_private;
	unsigned clusterpages = erofs_clusterpages(EROFS_SB(sb));
	struct z_erofs_vle_zipped_pack *victim = NULL;
	unsigned j, z_avail = 0; /* avoid the false uninitialized warning */
#endif
	unsigned i;
	struct bio_vec *bvec;

	bio_for_each_segment_all(bvec, bio, i) {
		struct page *page = bvec->bv_page;
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 13, 0))
		const int err = bio->bi_status;
#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 3, 0))
		const int err = bio->bi_error;
#endif
		struct z_erofs_vle_zipped_pack *zip;

		/* page is already locked */
		DBG_BUGON(PageUptodate(page));

		if (unlikely(err))
			SetPageError(page);
		else
			SetPageUptodate(page);

		debugln("%s: %d zpage %p index: %lu", __func__, __LINE__,
			page, page->index);

		zip = (void *)erofs_page_private(page);
		DBG_BUGON(zip == NULL);

		DBG_BUGON(!has_page_bundle(page));

#if EROFS_PAGE_BUNDLE_MAX_PAGES > 1
		/* for multiple bundle pages */
		if (zip == victim)
			++z_avail;
		else {
			z_avail = 0;
			for(j = 0; j < EROFS_PAGE_BUNDLE_MAX_PAGES; ++j)
				z_avail += PageUptodate(zip->bundle.pages[j]);
			victim = zip;
		}

		if (z_avail == clusterpages) {
#else
		if (PageUptodate(zip->bundle.pages[0])) {
#endif

			debugln("queue work %p zpage %p zip %p", &zip->work, page, zip);

			queue_work(z_erofs_workqueue, &zip->work);
		}

		unlock_page(page);
		/* page could be reclaimed now */
	}
	bio_put(bio);
}

#if (LINUX_VERSION_CODE < KERNEL_VERSION(4, 3, 0))
static inline void zipped_sync_read_endio(struct bio *bio, int err)
#else
static inline void zipped_sync_read_endio(struct bio *bio)
#endif
{
	unsigned i;
	struct bio_vec *bvec;

	bio_for_each_segment_all(bvec, bio, i) {
		struct page *page = bvec->bv_page;
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 13, 0))
		const int err = bio->bi_status;
#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 3, 0))
		const int err = bio->bi_error;
#endif

		/* page is already locked */
		DBG_BUGON(PageUptodate(page));

		if (unlikely(err))
			SetPageError(page);
		else
			SetPageUptodate(page);

		unlock_page(page);
		/* page could be reclaimed now */
	}
	bio_put(bio);
}

static struct bio *zipped_prepare_bio(struct super_block *sb,
	erofs_blk_t blkaddr, bool sync)
{
	/* FIXME, need optimise */
	struct bio *bio = bio_alloc(GFP_NOIO | __GFP_NOFAIL, BIO_MAX_PAGES);

	BUG_ON(bio == NULL);
	bio->bi_end_io = sync ? zipped_sync_read_endio :
	                        zipped_async_read_endio;
	bio_set_dev(bio, sb->s_bdev);
	bio->bi_private = sb;
#if (LINUX_VERSION_CODE < KERNEL_VERSION(3, 14, 0))
	bio->bi_sector = blkaddr << LOG_SECTORS_PER_BLOCK;
#else
	bio->bi_iter.bi_sector = blkaddr << LOG_SECTORS_PER_BLOCK;
#endif
	return bio;
}

static void __submit_bio(struct bio *bio, unsigned op, unsigned op_flags)
{
	bio_set_op_attrs(bio, op, op_flags);
#if (LINUX_VERSION_CODE < KERNEL_VERSION(4, 8, 0))
	submit_bio(0, bio);
#else
	submit_bio(bio);
#endif
}

static void z_erofs_vle_submit_all(struct super_block *sb,
                                   struct z_erofs_zipped_pagevec *vec,
                                   bool sync)
{
	struct page *page, *tmp;
	pgoff_t last_page;
	struct bio *bio = NULL;

	if (z_erofs_zipped_pagevec_empty(vec))
		return;

	/* should not be NULL */
	tmp = z_erofs_zipped_pagevec_pop(vec);
	do {
		pgoff_t current_page;

		page = tmp;
		current_page = page->index;

		/* could contain the pagevec itself, pop "tmp" in advance */
		tmp = z_erofs_zipped_pagevec_pop(vec);

		debugln("%s, found vec=%p page %p, index=%lu",
			__func__, vec, page, current_page);

		DBG_BUGON(!PageLocked(page));

		if (bio != NULL && last_page + 1 != page->index) {
submit_bio_retry:
			__submit_bio(bio, REQ_OP_READ, 0);
			bio = NULL;
		}

		if (bio == NULL)
			bio = zipped_prepare_bio(sb, current_page, sync);

		if (bio_add_page(bio, page, PAGE_SIZE, 0) < PAGE_SIZE)
			goto submit_bio_retry;

		last_page = current_page;
	} while (tmp != NULL);

	if (bio != NULL)
		__submit_bio(bio, REQ_OP_READ, 0);
}

static int z_erofs_vle_normalaccess_readpage(struct file *file,
                                             struct page *page)
{
	struct erofs_map_blocks_iter m_iter = {
		.map = {.m_llen = 0, .m_plen = 0},
		.mpage = NULL
	};
	struct z_erofs_vle_zipped_iter z_iter = { .zip = NULL };
	struct z_erofs_zipped_pagevec z_pvec = { .page = NULL };
	struct z_erofs_zipped_pack_collector collector = {
		.list = LIST_HEAD_INIT(collector.list),
		.sync = true
	};
	LIST_HEAD(pagepool);

	int err = z_erofs_vle_do_read_page(page, &z_pvec,
		&z_iter, &m_iter, &pagepool, &collector);

	if (z_iter.zip != NULL) {
		vle_zipped_iter_dispatch(&z_iter, &collector);
		vle_zipped_iter_end(&z_iter);
	}

	if (!err) {
		struct super_block *sb = page->mapping->host->i_sb;

		/* submit all compressed page in the forward order */
		z_erofs_vle_submit_all(sb, &z_pvec, true);
		/* unzip all collected compressed pages */
		vle_zipped_collected_unzip_all(sb, &collector.list);
	} else {
		errln("%s, failed to read, err [%d]", __func__, err);
		z_erofs_zipped_pagevec_end(&z_pvec);
	}

	if (m_iter.mpage != NULL)
		put_page(m_iter.mpage);

	/* clean up the remaining free pages */
	put_pages_list(&pagepool);
	return err;
}

static inline int __z_erofs_vle_normalaccess_readpages(
	struct file *filp,
	struct address_space *mapping,
	struct list_head *pages, unsigned nr_pages, bool sync)
{
	struct erofs_map_blocks_iter m_iter = {
		.map = {.m_llen = 0, .m_plen = 0},
		.mpage = NULL
	};
	struct z_erofs_vle_zipped_iter z_iter = { .zip = NULL };
	struct z_erofs_zipped_pagevec z_pvec = { .page = NULL };
	struct z_erofs_zipped_pack_collector collector = {
		.list = LIST_HEAD_INIT(collector.list),
		.sync = sync
	};
	struct super_block *sb = mapping->host->i_sb;
	LIST_HEAD(pagepool);

	gfp_t gfp = mapping_gfp_constraint(mapping, GFP_KERNEL);

	for (; nr_pages; --nr_pages) {
		/* traversal in reverse order */
		struct page *page = list_entry(pages->next, struct page, lru);

		prefetchw(&page->flags);
		list_del(&page->lru);

		if (add_to_page_cache_lru(page, mapping, page->index, gfp))
			list_add(&page->lru, &pagepool);
		else {
			int err = z_erofs_vle_do_read_page(page, &z_pvec,
				&z_iter, &m_iter, &pagepool, &collector);

			if (err) {
				errln("%s, readahead error at page %lu of nid %llu",
					__func__, page->index,
					EROFS_V(mapping->host)->nid);
			}
			put_page(page);
		}
	}

	if (z_iter.zip != NULL) {
		vle_zipped_iter_dispatch(&z_iter, &collector);
		vle_zipped_iter_end(&z_iter);
	}

	/* submit all compresssed page in the forward order */
	z_erofs_vle_submit_all(sb, &z_pvec, sync);

	if (!sync)
		/* queue all collected compressed pages (ready) for workers */
		vle_zipped_collected_enqueue_all(&collector.list);
	else
		/* unzip all collected compressed pages */
		vle_zipped_collected_unzip_all(sb, &collector.list);

	if (m_iter.mpage != NULL)
		put_page(m_iter.mpage);

	/* clean up the remaining free pages */
	put_pages_list(&pagepool);
	return 0;
}

static int z_erofs_vle_normalaccess_readpages(
	struct file *filp,
	struct address_space *mapping,
	struct list_head *pages, unsigned nr_pages)
{
	return __z_erofs_vle_normalaccess_readpages(filp,
		mapping, pages, nr_pages,
		nr_pages < 4 /* sync */);
}

/* for uncompressed (aligned) files and raw access for other files */
const struct address_space_operations z_erofs_vle_normal_access_aops = {
	.readpage = z_erofs_vle_normalaccess_readpage,
	.readpages = z_erofs_vle_normalaccess_readpages,
};

