// SPDX-License-Identifier: GPL-2.0+
/*
 * linux/fs/erofs/unzip_generic.c
 *
 * Copyright (c) 2018 HUAWEI, Inc.
 *             http://www.huawei.com/
 * Created by Gao Xiang <gaoxiang25@huawei.com>
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file COPYING in the main directory of the Linux
 * distribution for more details.
 */
#include "unzip.h"

/*
 * - The temporarily workaround approach -
 *
 * In the future, we attempt to add new interfaces
 * which can support a customized vmap iterator
 * to avoid memory allocation.
 */
static DEFINE_MUTEX(pagemap_lock);
static struct page *pagemap_pages_global[2048];

void z_erofs_plain_copy(struct z_erofs_pack_info *pack,
                        struct page **in_pages,
                        unsigned nr_inpages,
                        unsigned long long lofs)
{
	void *src = NULL;
	struct z_erofs_de_pagevec d_pvec;
	struct page *out_pages[96];
	unsigned i, nr_pages, startnr, pageofs;

	DBG_BUGON(!pack->nr_pages);
	DBG_BUGON(!pack->queued_pages);

	pageofs = lofs & ~PAGE_MASK;
	startnr = lofs >> PAGE_SHIFT;

	nr_pages = pack->nr_pages;
	for(i = 0; i < nr_pages; ++i)
		out_pages[i] = NULL;

	z_erofs_de_pagevec_init(&d_pvec, pack->pages, 0);
	for(i = 0; i < pack->queued_pages; ++i) {
		struct page *page = z_erofs_de_pagevec_dequeue(&d_pvec);
		unsigned pagenr;

		DBG_BUGON(page == NULL);
		pagenr = page->index - startnr;

		DBG_BUGON(!PageLocked(page));
		DBG_BUGON(pagenr >= nr_pages);
		out_pages[pagenr] = page;
	}
	z_erofs_de_pagevec_end(&d_pvec, true);
	/* clear queued logical pages */
	pack->queued_pages = 0;

	for(i = 0; i < nr_pages; ++i) {
		struct page *page = out_pages[i];
		void *dst;
		const unsigned righthalf = PAGE_SIZE - pageofs;

		if (page == NULL) {
			if (src != NULL)
				kunmap_atomic(src);
			src = kmap_atomic(in_pages[i]);
			continue;
		}

		dst = kmap_atomic(page);
		if (src != NULL) {
			memcpy(dst, src + righthalf, pageofs);
			kunmap_atomic(src);
		}

		if (i >= nr_inpages) {
			kunmap_atomic(dst);
			z_erofs_onlinepage_endio(page);
			return;
		}

		src = kmap_atomic(in_pages[i]);
		memcpy(dst + pageofs, src, righthalf);
		kunmap_atomic(dst);
		z_erofs_onlinepage_endio(page);
	}
	kunmap_atomic(src);
}

struct unzip_fast_percpu_buffer {
	char data[PAGE_SIZE * DE_PAGEVEC_INLINE_ENTRYS];
};

static inline int unzip_fast_percpu(struct z_erofs_pack_info *pack,
                                    void *in, size_t inlen,
                                    unsigned long long lofs,
                                    unsigned llen,
                                    struct page **pagemap)
{
	static struct unzip_fast_percpu_buffer pcpubuf[NR_CPUS];
	unsigned nr_pages = pack->nr_pages;
	unsigned i, j, pageofs;
	void *addr;
	size_t outlen;
	int ret;

	if (nr_pages > DE_PAGEVEC_INLINE_ENTRYS)
		return -ENOTSUPP;

	/* should be auto-unwrapped */
	for(i = 0; i < DE_PAGEVEC_INLINE_ENTRYS; ++i)
		pagemap[i] = NULL;

	j = lofs / PAGE_SIZE;
	for(i = 0; i < pack->queued_pages; ++i) {
		struct page *const page = pack->pages[i];

		DBG_BUGON(page == NULL);
		DBG_BUGON(page->index - j >= nr_pages);
		pagemap[page->index - j] = page;
	}

	/* clear queued logical pages */
	pack->queued_pages = 0;

	addr = pagemap[0] == NULL || PageHighMem(pagemap[0]) ?
	       NULL : page_address(pagemap[0]);

	for(i = 1; i < nr_pages; ++i) {
		if (pagemap[i] == NULL ||
		    addr + PAGE_SIZE != page_address(pagemap[i]) ||
		    PageHighMem(pagemap[i])) {
			addr = NULL;
			break;
		}
		addr += PAGE_SIZE;
	}

	pageofs = lofs & ~PAGE_MASK;
	outlen = min((nr_pages << PAGE_SHIFT) - pageofs, llen);
	if (addr != NULL) {
		/* we are lucky, do in-place decompression */
		ret = erofs_unzip_lz4(in,
		                      page_address(pagemap[0]) + pageofs,
		                      inlen, outlen);
		for(i = 0; i < nr_pages; ++i) {
			DBG_BUGON(pagemap[i] == NULL);
			if (ret < 0)
				SetPageError(pagemap[i]);
			z_erofs_onlinepage_endio(pagemap[i]);
		}
	} else {
		preempt_disable();
		addr = pcpubuf[smp_processor_id()].data;

		ret = erofs_unzip_lz4(in, addr + pageofs, inlen, outlen);
		if (ret >= 0)
			outlen = ret;

		for(i = 0; i < nr_pages; ++i) {
			j = min(PAGE_SIZE - pageofs, outlen);
			if (pagemap[i] != NULL) {
				if (ret < 0)
					SetPageError(pagemap[i]);
				else {
					void *dst = kmap_atomic(pagemap[i]);

					memcpy(dst + pageofs, addr + pageofs, j);
					kunmap_atomic(dst);
				}
				z_erofs_onlinepage_endio(pagemap[i]);
			}
			addr += PAGE_SIZE;
			outlen -= j;
			pageofs = 0;
		}

		preempt_enable();
	}
	return ret;
}

int z_erofs_unzip_generic(struct z_erofs_pack_info *pack,
                          void *in, size_t inlen,
                          unsigned long long lofs,
                          unsigned llen)
{
	struct z_erofs_de_pagevec d_pvec;
	unsigned i, startnr, pageofs, nr_pages;
	struct page *pagemap_pages_inline[96], **pagemap_pages;

	void *addr;
	size_t outlen;
	int ret;

	DBG_BUGON(!pack->queued_pages);

	erofs_dbg_might_sleep();

	nr_pages = pack->nr_pages;
	if (nr_pages > 96) {
		pagemap_pages = pagemap_pages_global;
		mutex_lock(&pagemap_lock);
	} else {
		pagemap_pages = pagemap_pages_inline;
		ret = unzip_fast_percpu(pack, in, inlen,
		                        lofs, llen,
		                        pagemap_pages);
		if (ret != -ENOTSUPP)
			return ret;
	}

	for(i = 0; i < nr_pages; ++i)
		pagemap_pages[i] = NULL;

	startnr = lofs / PAGE_SIZE;

	debugln("%s, pack = %p queued_pages = %u",
		__func__, pack, pack->queued_pages);

	z_erofs_de_pagevec_init(&d_pvec, pack->pages, 0);
	for(i = 0; i < pack->queued_pages; ++i) {
		struct page *page = z_erofs_de_pagevec_dequeue(&d_pvec);
		unsigned pagenr;

		DBG_BUGON(page == NULL);

		pagenr = page->index - startnr;
		debugln("%s, pack = %p page = %p, index = %lu, pagenr = %u",
			__func__, pack, page, page->index, pagenr);

		DBG_BUGON(!PageLocked(page));
		DBG_BUGON(pagenr >= nr_pages);
		pagemap_pages[pagenr] = page;
	}
	z_erofs_de_pagevec_end(&d_pvec, true);

	/* clear queued logical pages */
	pack->queued_pages = 0;

	/*
	 * for preload or small compression-ratio files,
	 * it is very likely queued_pages == nr_pages.
	 */
	if (i != nr_pages) {
		for(i = 0; i < nr_pages; ++i) {
			if (pagemap_pages[i] == NULL)
				/* TODO! temporary work */
				pagemap_pages[i] = alloc_page(GFP_KERNEL | __GFP_NOFAIL);
		}
	}

	addr = erofs_vmap(pagemap_pages, nr_pages);
	/* FIXME! error handling case */
	BUG_ON(addr == NULL);

	pageofs = lofs & ~PAGE_MASK;
	outlen = min((nr_pages << PAGE_SHIFT) - pageofs, llen);

	debugln("%s, in %p out %p inlen %lu outlen %lu", __func__,
		in, addr + pageofs, inlen, outlen);

#if 0
	print_hex_dump(KERN_DEBUG, "raw data: ", DUMP_PREFIX_OFFSET,
		16, 1, in, inlen, true);
#endif

	/* we only support LZ4 currently */
	ret = erofs_unzip_lz4(in, addr + pageofs, inlen, outlen);
	if (ret > 0) {
		outlen = ret;
		ret = 0;
	}

	erofs_vunmap(addr, nr_pages);

	for(i = 0; i < nr_pages; ++i) {
		if (!PageLocked(pagemap_pages[i]))
			/* TODO! temporary work */
			__free_page(pagemap_pages[i]);
		else {
			if (ret)
				SetPageError(pagemap_pages[i]);
			z_erofs_onlinepage_endio(pagemap_pages[i]);
		}
	}

	if (pagemap_pages == pagemap_pages_global)
		mutex_unlock(&pagemap_lock);
	return ret;
}

